      program vibron_3atomic_molecules
c                                     written by R.Lemus, N.Manini and S.Oss
c                                     last revision: Nicola Manini - Dec. 2001
c			http://alpha.science.unitn.it/~oss/vibr3at.html
c     works "as is" if linked with standard lapack & nag libraries
c     avaliable at http://www.netlib.org/lapack/
c
c     Example of basic compilation line:
c
c      f77 -o vibr3at vibr3at.f racah.f $libdir/liblapack.a $libdir/libblas.a
c
c     where $libdir is whatever directory where you put the compiled
c     lapack and blas libraries
c     if nag or imsl libraries are available/preferred: please look for LIBRARY
c     in the code and comment the lapack lines and uncomment the other
c     libraries lines 

      implicit real*8(a-h,o-z)

      integer dim1,dim2,dim3,maxp,muxp,muxps,muxnd,muxnds
      parameter(dim1=100)  ! dimension of "small" arrays
      parameter(dim2=204)  ! maximum # of data
      parameter(dim3=325)  ! maximum dimension of a multiplet
      parameter(maxp =20)  ! maximum number of parameters you can use
      parameter(muxp =18)  ! number of diag. parameters at disposal
      parameter(muxps=13)  ! number of diag. param. at disposal in symm. case
      parameter(muxnd= 1)  ! number of linear non diagonal operators
      parameter(muxnds=1)  ! number of linear non diag. operators in symm.case
 
      logical done,last,smetry,larity,wghted,iffit(maxp),ifcalc(muxnd)
      
      character*2  nov1,nov2
      character*12 stri1,stri2
      character*15 sto,stn,namn(muxnd),name(muxp)
      character*30 operd(maxp),opern(maxp)
      character*80 remark
      integer   n1,n2,nfam,v1,v2,vb,ifail
      integer   index(dim3),nq(2,dim3)
      integer   indexd(maxp),indexn(maxp)
      integer   ls(dim1),nqt(dim1),l2(dim1),nband(dim1)
      integer   lv(3,dim2),nlev(dim2)
      real*8    param(maxp)
      real*8    ceffe(maxp,maxp),adef(maxp)
      real*8    effe(maxp,dim3),apro(maxp,dim3),maxcomp
      real*8    eexpr(2,dim2),enert(dim3)
      real*8    hpr(maxp,dim3,dim3)
      real*8    scratch(muxp*dim3),scratch1(muxnd*dim3*dim3)
      real*8    ve(dim3,dim3),r(dim3),psi(dim3)
      real*8    h(dim3*dim3)  ! necessary for IMSL and for LAPACK
 
      done=.false.
      last=.false.

      call FACTRL(10)     !inizialize the factorials for the Racah coefficients

      print *,' loading parameters from file'
c                          the program starts inputing infos from default file:
      open(unit=5,file='tinp.dat',status='old')
c                      please comment previous line if you prefer reading stdin
      read(5,*) remark
      read(5,*) symm
      read(5,*) line
      read(5,*) weig            !inputing the global characters of the molecule

      smetry=(symm.gt.0)
      larity=(line.gt.0)
      wghted=(weig.gt.0)        !  calculating the global logical variables

      read(5,*) toll   !    inputing the final tolerance of the iterative fit
 
      if(smetry)then
         read(5,*) n1
         n2=n1
         nov1='vs'
         nov2='va'
      else
         read(5,*) n1,n2
         nov1='v1'
         nov2='v2'
      endif
c                           inputing the parameters fixing the representations
c                           of the total u(4) algebras
 
      if (smetry) then
         open(unit=3,file='vibr3atns.dat',status='old')
         much=muxps
         mucn=muxnds
      else
         open(unit=3,file='vibr3atnn.dat',status='old')
         much=muxp
         mucn=muxnd
      endif
      do j=1,much
         read(3,*) name(j)
      end do
      do j=1,mucn
         read(3,*) namn(j)
      end do
      close(3)
 
      ntruepar=0
      kpar=0
      do j=1,muxnd
         ifcalc(j)=.false.
      end do
 
1050  continue
      read(5,*) paro,iffi,sto,stn
      if(((paro.ne.99).or.(iffi.ne.99)
     &   .or.(sto.ne.'end')).and.(kpar.le.maxp)) then
         kpar=kpar+1
         param(kpar)=paro
         iffit(kpar)=(iffi.gt.0)
         ntruepar=ntruepar+iev(iffit(kpar))
         operd(kpar)=sto
         opern(kpar)=stn
         if (' '.eq.sto) then
            indexd(kpar)=0
         else
            indexd(kpar)=-1
            do i=1,much
               if (name(i).eq.sto) then
                  indexd(kpar)=i !  the diagonal operator has been recognized
               endif
            end do
         end if
         if(indexd(kpar).eq.-1)then
            print *,sto,': unknown name of diagonal operator ',kpar
            stop
         end if
         if (' '.eq.stn) then
            indexn(kpar)=0
         else
            indexn(kpar)=-1
            do i=1,mucn
               if (namn(i).eq.stn) then
                  if(i.eq.1)then  
                     indexmaj=kpar 
                  end if
                  indexn(kpar)=i !the not diagonal operator has been recognized
                  ifcalc(i)=.true.
               endif
            end do
         end if
         if(indexn(kpar).eq.-1) then !case no operator name
            print *,stn,': unknown name of not diagonal operator',kpar
            stop
         end if
         goto 1050              !                    repeat input loop
      end if
 
      print *,kpar,' operators in hamiltonian'
      print *,ntruepar,' fitted parameters in hamiltonian'
      
      write(*,'(/,a80)') remark
      if(wghted)then
         print *
         print *,' Weighted fit'
      endif
      print *
      print *,' Now loading energy data from file'
      
      lff=0
      lyy=0
      
      if (larity) then
         read(5,*) nfam
c                                       nfam is the number of species of bands:
c                                       for example nfam=2 means you have
c                                       in the input file states with 2 dif-   
c                                       ferent values of l, maybe some 
c                                       sigma and some delta band
      else
         nfam=1
      endif
      do iff=1,nfam
         read(5,*) nband(iff)   ! nband is the number of multiplets of this specie
         if(larity)then
            read(5,*) l2(iff)   ! l2 is l, i.e. the species
            if(abs(l2(iff)).gt.9) then ! little check
               print *,' l > 9 : there must be some mistake in data!!'
               stop
            endif
         else
            l2(iff)=0         ! l2 is k, in the bent case, and it is always =0
         endif
         do jff=1,nband(iff)
            lyy=lyy+1
            read(5,*) nqt(lyy),ls(lyy)
c                                nqt is the total number of (bent molecule) 
c                                   quanta in this multiplet
c                                ls is the number of states in this multiplet
            do kc=1,ls(lyy)
               lff=lff+1        ! lff is the experimental-state index
               if(wghted)then
                  read(5,*) nlev(lff),lv(1,lff),lv(2,lff),lv(3,lff),
     &                 eexpr(1,lff),eexpr(2,lff)
c                            nlev is the position of the state in the multiplet
c                            lv(i, ) is the experimental assignement
c                                    of the state (not used in the calculation)
c                            eexpr(1, ) is the experimental energy of the state
c                            eexpr(2, ) is the experimental error on the energy
c                                       (to be inserted in wghted fits only)
               else
                  read(5,*) nlev(lff),lv(1,lff),lv(2,lff),lv(3,lff),
     &                 eexpr(1,lff)
                  eexpr(2,lff)=1
               endif
               eexpr(2,lff)=eexpr(2,lff)**2 ! it is a least squares fits...
            end do
         end do
      end do
 
      istates=lff
 
      if(istates.gt.dim2) then  ! dimension check
         print *,' too many states to fit: increase dim2 in source!'
         stop
      endif
 
      print *,' Now fitting ',istates,' data with ',ntruepar,
     &     ' parameters '
 
      wnorm=1.
      olderr=1.d55
      oldolderr=1.d77
      
 9090 continue                  !   start of the iterative fit/calculation loop
      lff=0
      lyy=0
      
      if (.not.(last)) then
         print *,'________________________________________',
     &        '_________________________'
         if(wghted)then
            write(*,7654) dsqrt(sumsq/wnorm)
         end if
         write(*,7655) dsqrt(stde/(istates-1))
 7654    format(/,'            weighted rms deviation = ',f9.2)
 7655    format(/,' Standard deviation (not weighted) = ',f9.2,/)
         sumsq=0.
         wnorm=0.
         stde=0.
         do iu=1,kpar
            if(iffit(iu)) then
               write(*,7653) param(iu),operd(iu),opern(iu)
 7653          format(d14.6,3x,a30,1x,a30)
            endif
         end do
      endif                     ! (.not.(last))

 
      do iff=1,nfam             !   main loop on character of bands
         if (last) then
            write(3,*) 
     &'================================================================'
         end if
 
         ntau2= l2(iff)
 
         do jff=1,nband(iff)    !   main loop on bands of a fixed character
 
            lyy=lyy+1
            ntau1= n1+n2-l2(iff)-nqt(lyy)*2!explicit conversion of nqt to ntau1
 
            call basis(n1,n2,ntau1,ntau2,nq,numbst,dim3)
c     					nq and numbst are output

            call matel(n1,n2,hpr,kpar,numbst,nq,ntau1,ntau2,
     &           ifcalc,indexd,indexn,smetry,
     &           dim3,maxp,muxp,muxps,muxnd,muxnds,scratch,scratch1)
c                       calculates all the matrix elements of the algebraic
c                       operators, in this multiplet, putting them in hpr

            do i=1,numbst
c                       calculation of the algebraic hamiltonian matrix
               do j=1,numbst
                  ve(i,j)=0.
                  do ipar=1,kpar
                     ve(i,j)=ve(i,j)+param(ipar)*hpr(ipar,i,j)
                  end do
!                  h((i-1)*dim3+j)=ve(i,j)              ! IMSL
                  if(i.le.j) h(i + (j-1)*j/2)=ve(i,j)  ! LAPACK
               end do
            end do

! LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY !!!!
c                       diagonalization of h (the hamiltonian matrix, input)
c                       ve is the matrix of eigenvectors of h (output)
c                       r is the vector of eigenvalues (output)
c                       numbst and dim3 are the logical and physical dimensions
          call dspev('V','U',numbst,h,r,ve,dim3,scratch,ifail)	!	LAPACK

!            call devcsf(numbst,h,dim3,r,ve,dim3)		! IMSL
c     imsl v.2.0 puts the eigenvalues i descending order: please uncomment
c     the following 7 lines that re-sort in ascending order as required
cCc            do i=1,numbst/2			! IMSL
cCc               j=numbst-i+1			! IMSL
cCc               call swap(r(i),r(j))		! IMSL
cCc               do k=1,numbst			! IMSL
cCc                  call swap(ve(k,i),ve(k,j))	! IMSL
cCc               end do			! IMSL
cCc            end do				! IMSL

!          call f02faf('V','U',numbst,ve,dim3,r,scratch,muxp*dim3,ifail) ! NAG
c                       here the hamiltonian matrix (input) is in ve
c                       h is not used!
! LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY !!!!

            if (last) then      !       if true (final computation) generates
c                 the computed levels, if false it runs the fitting procedure
               write(3,*)'________________________________________',
     &              '______________________________'
               write(3,*)
     &  ' ___________________ Local basis ________________________'
               if (larity) then
                  write(3,1573) ntau1,ntau2
               else
                  write(3,1574) ntau1,ntau2
               endif
 1573          format('   #            v1  v2  vb^ l     Tau1=',
     &              i3,'  Tau2=',i3)
 1574          format('   #            v1  v2  vb        Tau1=',
     &              i3,'  Tau2=',i3)
 
               do ik=1,numbst   !  main loop on the local basis
                  v1=(n1-nq(1,ik))/2 !  conversion from algebraic omega
                  v2=(n2-nq(2,ik))/2 !  to traditional v quantum numbers
                  l=ntau2
                  
                  if (larity) then
                     vb=nq(1,ik)+nq(2,ik)-ntau1
                     write(3,1665) ik,v1,v2,vb,l
 1665                format(i4,10x,3i4,'^',i2)
                  else
                     vb=(nq(1,ik)+nq(2,ik)-ntau1-ntau2)/2
                     write(3,1666) ik,v1,v2,vb
 1666                format(i4,10x,3i4)
                  endif
               end do
 
               write(3,*)
     &  ' ________________ Diagonal basis ________________________'
               if (larity) then
                  write(3,6574) nov1,nov2
               else
                  write(3,6575) nov1,nov2
               endif
 6574          format('   #    energy  ',a2,2x,a2,
     &              '  vb^ l             components on local basis')
 6575          format('   #    energy  ',a2,2x,a2,
     &              '  vb                components on local basis')

               do jk=1,numbst !  main loop on h-eingenstates basis
                  maxcomp=0.d0
                  do jjk=1,numbst
                     abva=dabs(ve(jjk,jk))
                     if(abva.gt.maxcomp)maxcomp=abva
                  end do
                  if(sensitivity.gt.0.d0)then
                     iup=0
                     do jjk=1,numbst
                        abva=dabs(ve(jjk,jk))
                        if(abva.ge.sensitivity) then
                           iup=iup+1 !           selects the most significative
                           psi(iup)=ve(jjk,jk) ! components of the wavefunction
                           index(iup)=jjk !      to be printed
                        end if
                     end do
                  else
                     thre=maxcomp
 5432                continue
                     iup=0
                     do jjk=1,numbst
                        abva=dabs(ve(jjk,jk))
                        if(abva.ge.thre) then
                           iup=iup+1 !           selects the most significative
                           psi(iup)=ve(jjk,jk) ! components of the wavefunction
                           index(iup)=jjk !      to be printed
                        end if
                     end do
                     if(iup.le.1.and.thre.gt.0.05d0)then
                        thre=thre*0.8d0
                        goto 5432
                     end if
                  end if        !  (sensitivity.gt.0.d0)
                  parity=1
                  nmains=0
                  do jjk=1,numbst
                     if(dabs(ve(jjk,jk))/maxcomp.ge.0.999) then
                        imax=jjk ! selects the main component(s) of the state
                        nmains=nmains+1
                        parity=parity*ve(jjk,jk) ! used only in the symm. case
                     end if     ! to calculate normal vibrations
                  end do
 
                  v1=(n1-nq(1,imax))/2 ! local assignement of normal states
                  v2=(n2-nq(2,imax))/2
                  l=ntau2
 
                  if (smetry) then ! calculation of normal quantum
                     isum=v1+v2    ! numbers v1 and v2 (i.e. vs and va)
                     idif=abs(v1-v2)
                     if (parity.gt.0) then
                        ip=0
                     else
                        ip=1
                     endif
                     if(param(indexmaj).le.0.and.mod(isum,2).gt.0)then     
                        ip=1-ip
                     end if
                     if (nmains.gt.1) then
                        v1=idif-ip
                        v2=isum-idif+ip
                     endif
                     if (idif.eq.0) then               
                        v1=0
                        v2=isum
                     endif
                     if(param(indexmaj).le.0)then    
                        holdn=nva
                        nva=nvb
                        nvb=holdn            
                     end if          
                  endif! end of calculation of normal quantum numbers v1 and v2

                  jup=1
                  if (larity) then
                     vb=nq(1,imax)+nq(2,imax)-ntau1 ! vb in linear case
                     write(3,1765) jk,r(jk),v1,v2,vb,l,
     &                    (index(jup),psi(jup),jup=1,iup)
 1765                format(i4,f10.1,3i4,'^',i2,' | ',
     &                    1(0x,5(i4,f5.2)))
!     &                    1(<32*iev(jup.gt.5)>x,5(i4,f5.2)))
                  else          ! 			non linear case
                     vb=(nq(1,imax)+nq(2,imax)-ntau1-ntau2)/2
                     write(3,1766) jk,r(jk),v1,v2,vb,
     &                    (index(jup),psi(jup),jup=1,iup)
 1766                format(i4,f10.1,3i4,'    | ',
     &                    1(0x,5(i4,f5.2)))
!     &                    1(<32*iev(jup.gt.5)>x,5(i4,f5.2)))
                  endif
                  
               end do
            else                !      (if last)
c                                                        fitting part:
               do kc=1,ls(lyy)
                  lff=lff+1
                  ii=nlev(lff)
                  enert(lff)=r(ii)

                  sus=0.
                  do ios=1,numbst !		ve might not be unitary
                     sus=sus+ve(ios,ii)**2 ! so we need the norm of its columns
                  end do

                  do ipar=1,kpar
                     effep=0.
                     do ios=1,numbst ! effep is the contribution to the energy
                        do jos=1,numbst ! of state ii and given by the ipar
                           effep=effep+ ! parameter
     &                          ve(jos,ii)*ve(ios,ii)*hpr(ipar,jos,ios)
                        end do
                     end do
                     effe(ipar,lff)=effep/sus
c                         this is the normalization of the contribution
c                         to the calculated energy enert(lff)
c                         (of the lffth state) given by the iparth parameter
                     apro(ipar,lff)=eexpr(1,lff)*effe(ipar,lff)/
     &                    eexpr(2,lff)
                  end do

                  wnorm=wnorm+1.0/eexpr(2,lff)
                  sumsq=sumsq+(enert(lff)-eexpr(1,lff))**2/eexpr(2,lff)
                  stde=stde+(enert(lff)-eexpr(1,lff))**2
 
                  if (done) then !	at convergence: summary of fit result
                     write(3,123) nqt(lyy),l2(iff),nlev(lff),
     &                    lv(1,lff),lv(2,lff),lv(3,lff),
     &                    eexpr(1,lff),enert(lff),
     &                    enert(lff)-eexpr(1,lff)
                  end if
 123              format(6i4,f11.3,3f10.2)
               end do           ! kc loop
c                  the experimental states in the multiplet with fixed nqt(lyy)
            end if  ! if last - the last time one spares the fitting procedure 
         end do                 ! jff=1,nband(iff)
      end do                    ! iff=1,nfam  : main loop on character of bands
c            end of the loops on the multiplets with fixed v1+v2+vb and species
 
      if(.not.(last.or.done)) then

         do imu=1,kpar          !        calculation of the known vector
            adef(imu)=0.        !            in the least squares theory
            do ika=1,istates
               adef(imu)=adef(imu)+apro(imu,ika)
            end do
         end do

         do imu=1,kpar          !  calculation of the coefficient matrix
            do ini=1,kpar       !            in the least squares theory
               ceffe(imu,ini)=0.
               do ika=1,istates
                  ceffe(imu,ini)=ceffe(imu,ini)
     &                 +effe(imu,ika)*effe(ini,ika)/eexpr(2,ika)
               end do
            end do
         end do
 
         do io1=1,kpar
            if (.not.iffit(io1)) then
               do io2=1,kpar
                  ceffe(io1,io2)=0
               end do
               ceffe(io1,io1)=1.
               adef(io1)=param(io1)
            endif       ! does not change the parameters we decided not to fit
         end do

! LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY !!!!
c                               solution of linear system      ceffe*param=adef
c                         kpar and maxp are the logical and physical dimensions
         call DCOPY(maxp,adef,1,param,1)			!	LAPACK
         call dgesv(kpar,1,ceffe,maxp,scratch,param,maxp,ifail) !	LAPACK

!         call dlsarg(kpar,ceffe,maxp,adef,1,param)		!	IMSL

!         call f04arf(ceffe,maxp,adef,kpar,param,scratch,ifail) !	NAG

! LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY LIBRARY !!!!


         if ((dabs(sumsq-olderr).le.toll).or.
     &        (sumsq.ge.(oldolderr-0.01))) then
            if (sumsq.ge.(oldolderr-0.01)) then
               print *,' convergence troubles...'
            endif
            done=.true.
 
            open(unit=3,file='tout.dat',status='unknown')
c                                       this is the name of the output file
            write(3,*)
            write(3,*) ' ',remark
            write(3,776) istates,ntruepar
 776        format(//' Fit of',i4,' data with',i3,
     &           ' effective parameters.',///,' Statistic of the fit: ')
 
            if(wghted)then
               write(3,7654) dsqrt(sumsq/wnorm)
            end if
            write(3,7655) dsqrt(stde/(istates-1))
            write(3,778) n1,n2
 778        format(//' Final parameters: ',/,
     &           '                  n1=',i3,'   n2=',i3,/)
 
            do ip=1,kpar   !list of best fit parameters 
               if (iffit(ip)) then
                  stri1=' '
               else
                  stri1=' not fitted'
               endif
               if(operd(ip).ne.' '.and.opern(ip).ne.' ')then
                  stri2=' * '
               else
                  stri2=' '
               endif
               write(3,65) param(ip),operd(ip),stri2,opern(ip),stri1
            end do
 65         format(d15.8,1x,a25,a3,a25,a11)
            write(3,*)
            write(3,6675) nov1,nov2 !   head label of the summary of fit result
 6675       format(
     &' nvb  s2 pos  ',a2,2x,a2,'  vb       Eobs     Ecalc    deltaE ')
         end if
         oldolderr=olderr
         olderr=sumsq
         goto 9090              !            closes the iterative fitting loop
      end if                    !            if(.not.(last.or.done))

      if(.not.last)then
         write(3,*)
         done=.false.
         last=.true.
      
         print *
         print *,' computing levels according to input file.'
         read(5,*) sensitivity
      
         if(larity)then
            read(5,*) nfam
         else
            nfam=1
         endif
         do i=1,nfam
            read(5,*) nband(i)  ! reading which levels we want to be calculated
            print *,'# of bands:',nband(i)
         end do
         close(5)
 
         llz=0
         sts=0
         do i=1,nfam
            l2(i)=i-1
            sts=sts+(nband(i)**3+6*nband(i)**2+11*nband(i)+6)/6
            do j=1,nband(i)
               llz=llz+1
               nqt(llz)=j-1
               ls(llz)=j*(j+1)/2
            end do
         end do
         goto 9090              ! last iteration, for computing levels
      end if
      
      close(3)

c                        finally creates a useful file containing only the
c                        final parameter values, with the correct format,
c                        that may be included in the input
c                        file for a new, possibly extended, fit
      open(unit=8,file='parguess.dat',status='unknown')
      do iu=1,kpar
         integ=iev(iffit(iu))
         write(8,1435) param(iu),integ,
     &        char(39)//operd(iu)(1:lastblank(operd(iu)))//char(39),
     &        char(39)//opern(iu)(1:lastblank(opern(iu)))//char(39)
      end do
 1435 format(d14.6,i3,2x,a30,1x,a30)
 
      close(8)
 
9999  end     !          end of main
 
c______________________________________________________________________________
      subroutine basis(n1,n2,ntau1,ntau2,nq,numbst,dim3)
c            generates the states compatible with the values of ntau1,ntau2
 
      implicit none
      integer n1,n2,ntau1,ntau2,numbst,dim3,nquanta,na,nb
      integer v1,v2,vb,nq(2,dim3)

      numbst=0
      nquanta=(ntau2-ntau1+n1+n2)/2
c                             it's the number of quanta (in the "bent" sense)
c                             we have in this group of states
c                             with tau1 and tau2 being fixed.
 
c                             now we distribute them among v1,v2 and vb,
c                             with the algebraic condition that vb>=tau2
      do vb=nquanta,ntau2,-1
         do v1=0,nquanta-vb
            v2=nquanta-v1-vb
            na=n1-2*v1
            nb=n2-2*v2
            if((vb.le.na).and.(vb.le.nb)) then
c                             this condition has to be checked in order not
c                             to go outside the finite-dimensional space
c                             of our representation of u(4) x u(4);
c                             in fact, it's automatically true in the case
c                             that  nquanta<min(n1 /2,n2 /2)
               numbst=numbst+1
               if(numbst.gt.dim3) then ! dimension check
                  print *,'Error: multiplet too big:',
     &                 'please increase dim3 in source !'
                  stop
               endif
               nq(1,numbst)=na
               nq(2,numbst)=nb
            endif
         end do
      end do
      return
      end
 
c_____________________________________________________________________
 
      subroutine matel(n1,n2,hpr,kpar,numbst,nq,t1,t2,
     &     ifcalc,indexd,indexn,smetry,
     &     dim3,maxp,muxp,muxps,muxnd,muxnds,hdi,hndi)
c                       calculates the matrix elements of the algebraic
c                       operators, putting them into hpr

      implicit real*8(a-h,o-z)

      integer dim3,maxp,muxp,muxps,muxnd,muxnds
      integer n1,n2,kpar,w1,w2,wp1,wp2,t1,t2,n12
 
      logical smetry,ifcalc(muxnd)

      real*8 hdi(muxp,dim3),hndi(muxnd,dim3,dim3),hpr(maxp,dim3,dim3)
      integer indexd(maxp),indexn(maxp),nq(2,dim3)

      do 10 i=1,numbst
         do 10 j=1,numbst
           do 11 ii=1,kpar
11            hpr(ii,i,j)=0
           do 12 ii=1,muxnd
12            hndi(ii,i,j)=0
10    continue
 
      n12=n1+n2
      ak1=1.0*(n12+1)/(n1+1)
      ak2=1.0*(n12+1)/(n2+1)
      bk1=n1*(n1+2)
      bk2=n2*(n2+2)
      co12=t1*(t1+2)+t2**2-n12*(n12+2) !    the original 12 invariant
      cp12=dabs(1.d0*((t1+1)*t2))
 
      do 30 i=1,numbst
       w1=nq(1,i)
       w2=nq(2,i)
 
       c1=w1*(w1+2)-bk1
       c2=w2*(w2+2)-bk2
       c12=1.0*co12-ak1*c1-ak2*c2    !    the 12 invariant is "reduced"
 
c                      calculation of the diagonal matrix elements
       if(smetry)then                                       ! symmetrical case
 
        hdi(1,i)=c1+c2                ! linear invariants
        hdi(2,i)=c12
        hdi(3,i)=cp12
 
        hdi(4,i)=c1**2+c2**2          ! quadratic diagonal terms
        hdi(5,i)=c12**2
        hdi(6,i)=cp12**2
 
        hdi(7,i)=c1*c2                ! quadratic product terms
        hdi(8,i)=(c1+c2)*c12
        hdi(9,i)=(c1+c2)*cp12
        hdi(10,i)=c12*cp12
 
        hdi(11,i)=c1**3+c2**3         ! cubic diagonal terms
        hdi(12,i)=c12**3
        hdi(13,i)=cp12**3
 
       else                                              ! non-symmetrical case
 
        hdi(1,i)=c1                 ! linear invariants
        hdi(2,i)=c2
        hdi(3,i)=c12
        hdi(4,i)=cp12
 
        hdi(5,i)=c1**2              ! quadratic diagonal terms
        hdi(6,i)=c2**2
        hdi(7,i)=c12**2
        hdi(8,i)=cp12**2
 
        hdi(9,i)=c1*c2              ! quadratic product terms
        hdi(10,i)=c1*c12
        hdi(11,i)=c1*cp12
        hdi(12,i)=c2*c12
        hdi(13,i)=c2*cp12
        hdi(14,i)=c12*cp12
 
        hdi(15,i)=c1**3               ! cubic diagonal terms
        hdi(16,i)=c2**3
        hdi(17,i)=c12**3
        hdi(18,i)=cp12**3
       endif
 
c                           non-diagonal (majorana) matrix elements
        if (ifcalc(1)) then
         do 31 j=1,numbst
          wp1= nq(1,j)
          wp2= nq(2,j)
          hndi(1,i,j)=dmaj12(n1,n2,w1,w2,wp1,wp2,t1,t2)
   31    continue
        endif
 
30    continue
 
      do 50 k=1,kpar    !  storage of matrix elements in hpr
        in = indexd(k)
        do 50 i=1,numbst
         if(indexn(k).eq.0)then
          hpr(k,i,i)=hdi(in,i)
         else
          if(in.le.0)then
           do 61 j=1,numbst
61          hpr(k,i,j)=hndi(indexn(k),i,j)
          else
           do 60 j=1,numbst
60          hpr(k,i,j)=(hdi(in,i)+hdi(in,j))*hndi(indexn(k),i,j)
          endif
         endif
50    continue
      return
      end
 
c_______________________________________________________________________
 
      function dmaj12(n1,n2,w1,w2,wp1,wp2,t1,t2)
c     Calculates the matrix elements of the Majorana operator M12
c     between initial and final states (w1,w2) and (wp1,wp2) respectively
c     Requires a routine racahci(i1..i6) that computes the 
c     6-j symbol in the Wigner (not Racah!) phase convention
c     i1=2*j1... i6=2*j6 , where j1...j6 are the true angular momenta

      implicit none

      real*8 dmaj12,racahci,red
      integer n1,n2,w1,w2,wp1,wp2,t1,t2
 
      dmaj12=0.d0
      if( (wp1.eq.w1+2.or.wp1.eq.w1.or.wp1.eq.w1-2).and.
     &     (wp2.eq.w2+2.or.wp2.eq.w2.or.wp2.eq.w2-2) ) then
         dmaj12=( (-1)**(t1+1) )*
     &        (wp1+1)*(wp2+1)*
     &        racahci(w1,w2,wp1,wp2,(t1+t2),2)*
     &        racahci(w1,w2,wp1,wp2,(t1-t2),2)*
     &        red(n1,w1,wp1) * red(n2,w2,wp2)
      end if
      if(wp1.eq.w1.and.wp2.eq.w2)
     &     dmaj12 = dmaj12 - (1./4.)*( (t1*(t1+2)+t2**2) - 
     &     w1*(w1+2) -w2*(w2+2) ) + (1./2.)*n1*n2 -
     &     (-1)**(n1+n2+1)*(n1+1)*(n1+2)*(n2+1)*(n2+2)*
     &     racahci(n2,n2,n1,n1,2,n1+n2)**2/4.
      return
      end
 
c___________________________________________________________________
 
      function red(n,w,wp)
c                            reduced matrix element of Majorana operator
      implicit none
      real*8 red
      integer n,w,wp
 
      red=0.d0
      if(wp.eq.w) then
         red=(1.d0)*(n+2)/2.
         return
      end if
      if(wp.eq.w+2) then
         red = dsqrt(1.d0*(n-w)*(n+w+4)*(w+1)/(w+3))/2.
         return
      end if
      if(wp.eq.w-2) then
         red = dsqrt(1.d0*(n-w+2)*(n+w+2)*(w+1)/(w-1))/2.
         return
      end if
      return
      end
c___________________________________________________________________________

      function racahci(i1,i2,i3,i4,i5,i6) !	THREE VERSIONS
c     computes the Wigner 6-j coefficient as defined 
c     in eq. (15.3) of de-Shalit & Talmi, "Nuclear Shell Theory" (1963)
      implicit none
      real*8 racahci,racahc
      integer i1,i2,i3,i4,i5,i6

c     We assume the routines racahc and gracah for the Racah coefficients take
c     as input already  the integer variables 2*j1,2*j2,... , where j1,j2,...
c     are the true (integer or half-integer) angular momenta.

c     The gracah subroutine produces already the 6-j symbol as defined in
c     in eq. (15.3) of de-Shalit & Talmi, "Nuclear Shell Theory" (1963):
      call gracah(i1,i2,i3,i4,i5,i6,racahc)
      racahci=racahc

c     If we use instead a routine racahc computing the Racah 6-j symbol as in 
c     eq. (15.5) of de-Shalit & Talmi's book, then compatibility 
c     with the rest of the code (that uses the Wigner 6-j symbols)
c     needs transposing the arguments and adding a phase as in
c     eq. (15.8) of de-Shalit & Talmi's book
c      racahci=(-1)**((i1+i2+i3+i4)/2) *racahc(i1,i2,i5,i4,i3,i6)

c     This old version provided a 0.5 factor in the inputs of the 6-j 
c     coefficients for a function that takes real*8 true angular momenta as 
c     inputs:
c      racahci=racahc((0.5d0*i1),(0.5d0*i2),(0.5d0*i3),
c     &               (0.5d0*i4),(0.5d0*i5),(0.5d0*i6))

      return
      end
c___________________________________________________________________________

      function iev(log)  ! converts logical into integer variable

      implicit none
      integer iev
      logical log

      if (log) then 
         iev=1
      else
         iev=0
      end if

      return
      end   
c___________________________________________________________________________

      function lastblank(stringa)
c                       finds the last nonblank position of a character string
      implicit none
      integer i,lastblank
      character*(*) stringa            

      do i=len(stringa),2,-1        
         if(stringa(i:i).ne.' ') goto 21
      end do
 21   continue
      lastblank=i                      
      return
      end

c___________________________________________________________________________

      subroutine swap(x,y)
      implicit none
      real*8 x,y,tmp

      tmp=x
      x=y
      y=tmp
      return
      end
